use std::{collections::VecDeque, time::{Instant, Duration}, thread, fs::{self, File}, env, path::PathBuf, io::{Read, self, ErrorKind, Write}, process, fmt::Display};
use data::{TetriminoData, WallkickTable, WALLKICK_TABLE_1, WALLKICK_TABLE_2, Rotation};
use pancurses::{initscr, noecho, Input, endwin, start_color, init_color, use_default_colors, init_pair, COLOR_RED, COLOR_CYAN, COLOR_YELLOW, COLOR_MAGENTA, COLOR_GREEN, COLOR_BLUE, Window, COLOR_WHITE, curs_set, cbreak};
use rand::prelude::*;
use serde::de::{Visitor, self};
use serde_derive::{Deserialize, Serialize};

mod data;

#[derive(Debug, Clone, Copy)]
enum Color {
    Cyan, // I
    Yellow, // O
    Purple, // T
    Green, // S
    Red, // Z
    Blue, // J
    Orange // L
}

const COLOR_ORANGE: i16 = 8;

impl Color {
    fn into_color_pair(self) -> i16 {
        match self {
            Self::Cyan => COLOR_CYAN,
            Self::Yellow => COLOR_YELLOW,
            Self::Purple => COLOR_MAGENTA,
            Self::Green => COLOR_GREEN,
            Self::Red => COLOR_RED,
            Self::Blue => COLOR_BLUE,
            Self::Orange => COLOR_ORANGE
        }
    }
}

#[derive(Debug, Clone, Copy)]
enum TetriminoKind {
    I, O, T, S, Z, J, L
}

#[derive(Debug, Clone, Copy)]
struct Tetrimino {
    kind: TetriminoKind,
    rotation: Rotation
}

impl Into<Color> for Tetrimino {
    fn into(self) -> Color {
        match self.kind {
            TetriminoKind::I => Color::Cyan,
            TetriminoKind::O => Color::Yellow,
            TetriminoKind::T => Color::Purple,
            TetriminoKind::S => Color::Green,
            TetriminoKind::Z => Color::Red,
            TetriminoKind::J => Color::Blue,
            TetriminoKind::L => Color::Orange,
        }
    }
}

impl Into<TetriminoData> for Tetrimino {
    fn into(self) -> TetriminoData {
        let i = self.rotation as usize;
        match self.kind {
            TetriminoKind::I => data::I[i],
            TetriminoKind::O => data::O[i],
            TetriminoKind::T => data::T[i],
            TetriminoKind::S => data::S[i],
            TetriminoKind::Z => data::Z[i],
            TetriminoKind::J => data::J[i],
            TetriminoKind::L => data::L[i],
        }
    }
}

impl Tetrimino {
    fn data(self) -> TetriminoData {
        self.into()
    }

    fn color(self) -> Color {
        self.into()
    }

    fn wallkick_table(self) -> WallkickTable {
        match self.kind {
            TetriminoKind::I => WALLKICK_TABLE_2,
            _ => WALLKICK_TABLE_1
        }
    }

    fn rotate_cw(&mut self) {
        use Rotation::*;
        self.rotation = match self.rotation {
            Spawn => Clockwise,
            Clockwise => Flipped,
            Flipped => CounterClockwise,
            CounterClockwise => Spawn,
        }
    }

    fn rotate_ccw(&mut self) {
        use Rotation::*;
        self.rotation = match self.rotation {
            Spawn => CounterClockwise,
            Clockwise => Spawn,
            Flipped => Clockwise,
            CounterClockwise => Flipped,
        }
    }
}

struct TetriminoGenerator {
    remaining: VecDeque<Tetrimino>
}

impl TetriminoGenerator {
    fn new() -> Self {
        Self { remaining: VecDeque::new() }
    }
    
    fn reserve(&mut self, n: usize) {
        use TetriminoKind::*;
        while self.remaining.len() < n {
            let mut next_bag = [I, O, T, S, Z, J, L];
            next_bag.shuffle(&mut rand::thread_rng());
            self.remaining.extend(next_bag.into_iter().map(|v| Tetrimino {
                kind: v,
                rotation: Rotation::Spawn
            }));
        }
    }

    fn remaining(&mut self) -> &[Tetrimino] {
        self.remaining.make_contiguous();
        self.remaining.as_slices().0
    }
}

impl Iterator for TetriminoGenerator {
    type Item = Tetrimino;

    fn next(&mut self) -> Option<Self::Item> {
        self.reserve(1);

        // kinda useless to unwrap but its better to panic than to run out of tetriminoes
        Some(self.remaining.pop_front().unwrap())
    }
}

fn init_colors() {
    start_color();
    use_default_colors();
    init_pair(COLOR_CYAN, COLOR_CYAN, -1);
    init_pair(COLOR_YELLOW, COLOR_YELLOW, -1);
    init_pair(COLOR_MAGENTA, COLOR_MAGENTA, -1);
    init_pair(COLOR_GREEN, COLOR_GREEN, -1);
    init_pair(COLOR_RED, COLOR_RED, -1);
    init_pair(COLOR_BLUE, COLOR_BLUE, -1);
    init_color(COLOR_ORANGE, 1000, 843, 0);
    init_pair(COLOR_ORANGE, COLOR_ORANGE, -1);
    init_pair(COLOR_WHITE, COLOR_WHITE, -1);
}

fn write_tetrimino(t: Tetrimino, window: &Window, mut x: i32, mut y: i32) {
    window.color_set(t.color().into_color_pair());
    for row in t.data() {
        for cell in *row {
            if y >= 0 && *cell {
                window.mvprintw(y, x, "██");
            }
            x += 2;
        }
        x -= row.len() as i32 * 2;
        y += 1;
    }
}

fn write_tetrimino_centered(t: Tetrimino, window: &Window, x: i32, y: i32) {
    write_tetrimino(t, window, x + (t.data()[0].len() % 2) as i32, y)
}

fn write_tetrimino_ghost(t: TetriminoData, window: &Window, mut x: i32, mut y: i32) {
    for row in t {
        for cell in *row {
            if y >= 0 && *cell {
                window.mvprintw(y, x, "░░");
            }
            x += 2;
        }
        x -= row.len() as i32 * 2;
        y += 1;
    }
}

fn print_board(board: [[Option::<Color>; 10]; 40], window: &Window, x: i32, y: i32) {
    for (y2, row) in board.iter().enumerate() {
        for (x2, cell) in row.iter().enumerate() {
            if y2 as i32 + y < 0 { // moment
                continue;
            }
            window.mv(y2 as i32 + y, 2 * x2 as i32 + x);
            if let Some(cell) = cell {
                window.color_set(cell.into_color_pair());
                window.printw("██");
            } else {
                window.printw("  ");
            }
        }
    }
}

fn check_tetrimino_valid(t: TetriminoData, board: [[Option::<Color>; 10]; 40], mut x: i32, mut y: i32) -> bool {
    for row in t {
        for cell in *row {
            if *cell {
                if x > 9 || x < 0 || y < 0 || y > 39 || board[y as usize][x as usize].is_some() {
                    return false
                }
            }
            x += 1;
        }
        x -= row.len() as i32 * 1;
        y += 1;
    }

    true
}

fn lock_tetrimino(tetrimino: Tetrimino, board: &mut [[Option::<Color>; 10]; 40], mut x: i32, mut y: i32) {
    for row in tetrimino.data() {
        for cell in *row {
            if *cell {
                board[y as usize][x as usize] = Some(tetrimino.color());
            }
            x += 1;
        }
        x -= row.len() as i32 * 1;
        y += 1;
    }
}

#[derive(Serialize, Deserialize, Debug, Clone)]
struct Binds {
    cw_rot: Vec<Bind>,
    ccw_rot: Vec<Bind>,
    hard_drop: Vec<Bind>,
    soft_drop: Vec<Bind>,
    hold: Vec<Bind>,
    pause: Vec<Bind>,
    left: Vec<Bind>,
    right: Vec<Bind>,
    quit: Vec<Bind>,
}

#[derive(Debug, Clone, Copy)]
enum Bind {
    Char(char),
    UpArrow,
    DownArrow,
    LeftArrow,
    RightArrow,
}

fn display_bind_vec(bind: &Vec<Bind>) -> String {
    if let Some(bind) = bind.get(0) {
        format!("{}", bind)
    } else {
        "none".to_string()
    }
}

impl Display for Bind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            Bind::Char(c) => write!(f, "'{}'", c),
            Bind::UpArrow => write!(f, " ↑ "),
            Bind::DownArrow => write!(f, " ↓ "),
            Bind::LeftArrow => write!(f, " ← "),
            Bind::RightArrow => write!(f, " → "),
        }
    }
}

impl<'de> serde::Deserialize<'de> for Bind {
    fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
    where
        D: serde::Deserializer<'de> {
        deserializer.deserialize_str(BindVisitor)
    }
}

struct BindVisitor;

impl<'de> Visitor<'de> for BindVisitor {
    type Value = Bind;

    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
        // doesnt appear this is ever called, still implementing it
        write!(formatter, r#"expecting a char or one of "up", "down", "left", "right""#)
    }

    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
    where
        E: de::Error,
    {
        if v.chars().count() == 1 {
            Ok(Bind::Char(v.chars().next().unwrap()))
        } else {
            match v {
                "up" => Ok(Bind::UpArrow),
                "down" => Ok(Bind::DownArrow),
                "left" => Ok(Bind::LeftArrow),
                "right" => Ok(Bind::RightArrow),
                _ => Err(de::Error::unknown_variant(v, &["<char>", "up", "down", "left", "right"]))
            }   
        }
    }
}

impl serde::Serialize for Bind {
    fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
    where
        S: serde::Serializer {
        let s;
        serializer.serialize_str(match self {
            Bind::Char(c) => {
                s = c.to_string();
                &s
            },
            Bind::UpArrow => "up",
            Bind::DownArrow => "down",
            Bind::LeftArrow => "left",
            Bind::RightArrow => "right",
        })
    }
}

impl Default for Binds {
    fn default() -> Self {
        Self {
            cw_rot: vec![Bind::Char('x'), Bind::UpArrow],
            ccw_rot: vec![Bind::Char('z')],
            hard_drop: vec![Bind::Char(' ')],
            soft_drop: vec![Bind::DownArrow],
            hold: vec![Bind::Char('c')],
            pause: vec![Bind::Char('`')],
            left: vec![Bind::LeftArrow],
            right: vec![Bind::RightArrow],
            quit: vec![Bind::Char('q')]
        }
    }
}

impl PartialEq<Input> for Bind {
    fn eq(&self, other: &Input) -> bool {
        match (self, other) {
            (Self::Char(l), Input::Character(f)) => l == f,
            (Self::UpArrow, Input::KeyUp) => true,
            (Self::DownArrow, Input::KeyDown) => true,
            (Self::LeftArrow, Input::KeyLeft) => true,
            (Self::RightArrow, Input::KeyRight) => true,
            _ => false
        }
    }
}

fn get_binds() -> io::Result<Binds> {
    let home = env::var("HOME");
    if let Ok(home) = home {
        let mut path = PathBuf::from(home);
        path.push(".config");
        path.push("tetris");
        fs::create_dir_all(&path)?;
        path.push("binds.toml");
        let mut file = File::open(&path).or_else(|e| {
            if e.kind() == ErrorKind::NotFound {
                let mut file = File::create(&path)?;
                file.write_all(toml::to_string_pretty(&Binds::default()).unwrap().as_bytes())?;
                File::open(&path)
            } else {
                Err(e)
            }
        })?;
        let mut contents = String::new();
        file.read_to_string(&mut contents)?;
        Ok(toml::from_str(&contents).unwrap_or_else(|e| {
            eprintln!("Error parsing `{}`:\n{}", path.to_str().unwrap(), e);
            process::exit(1);
        }))
    } else {
        Ok(Default::default())
    }
}

fn main() {
    let binds = get_binds().unwrap_or(Default::default());

    let mut board = [[None; 10]; 40];

    let window = initscr();
    
    // dont put this before initscr, you get a segfault :)
    // "rust is memory safe" not if youre the ncurses crate author
    noecho();
    
    curs_set(0);

    cbreak();

    init_colors();

    window.keypad(true);

    window.nodelay(true);

    window.refresh();

    let mut rando = TetriminoGenerator::new();

    let mut hold = None;

    macro_rules! reset {
        ($a:ident, $b:ident, $c:ident, $d:ident, $e:ident, $f:ident, $g:ident) => {
            $a = Instant::now();
            $b = $f.next().unwrap();
            $c = 18;
            $d = 3;
            $e = false;
            if !check_tetrimino_valid($b.into(), $g, $d, $c) {
                endwin();
                process::exit(0)
            }
        }
    }

    let mut next_fall;
    let mut tetrimino;
    let mut y;
    let mut x;
    let mut hold_used;

    reset!(next_fall, tetrimino, y, x, hold_used, rando, board);

    'main: loop {
        let mut needs_refresh = false;

        let key = window.getch();

        if let Some(key) = key {
            if binds.pause.iter().any(|v| *v == key) {
                let before_pause = Instant::now();
                window.mvprintw(3,  12,         "⎛‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾‾⎞");
                window.mvprintw(4,  12,         "⎥      Paused      ⎥");
                window.mvprintw(5,  12,         "⎥                  ⎥");
                window.mvprintw(6,  12,         "⎥  -- Controls --  ⎥");
                window.mvprintw(7,  12, format!("⎥ {:>4} to rot cw   ⎥", display_bind_vec(&binds.cw_rot)));
                window.mvprintw(8,  12, format!("⎥ {:>4} to rot ccw  ⎥", display_bind_vec(&binds.ccw_rot)));
                window.mvprintw(9,  12, format!("⎥ {:>4} to hard drp ⎥", display_bind_vec(&binds.hard_drop)));
                window.mvprintw(10, 12, format!("⎥ {:>4} to soft drp ⎥", display_bind_vec(&binds.soft_drop)));
                window.mvprintw(11, 12, format!("⎥ {:>4} to hold     ⎥", display_bind_vec(&binds.hold)));
                window.mvprintw(12, 12, format!("⎥ {:>4} to unpause  ⎥", display_bind_vec(&binds.pause)));
                window.mvprintw(13, 12, format!("⎥ {:>4} to mv left  ⎥", display_bind_vec(&binds.left)));
                window.mvprintw(14, 12, format!("⎥ {:>4} to mv right ⎥", display_bind_vec(&binds.right)));
                window.mvprintw(15, 12, format!("⎥ {:>4} to quit     ⎥", display_bind_vec(&binds.quit)));
                window.mvprintw(16, 12,         "⎝__________________⎠");
                loop {
                    match window.getch() { 
                         Some(c) if binds.pause.iter().any(|v| *v == c) => break,
                        Some(c) if binds.quit.iter().any(|v| *v == c) => break 'main,
                        _ => (),
                    }
                }
                needs_refresh = true;
                let after_pause = Instant::now();
                let pause_len = after_pause - before_pause;
                next_fall += pause_len; // hacky but who cares
            }

            if binds.quit.iter().any(|v| *v == key) {
                break
            }

            if binds.left.iter().any(|v| *v == key) {
                if check_tetrimino_valid(tetrimino.into(), board, x - 1, y) {
                    needs_refresh = true;
                    x -= 1;
                }
            }

            if binds.right.iter().any(|v| *v == key) {
                if check_tetrimino_valid(tetrimino.into(), board, x + 1, y) {
                    needs_refresh = true;
                    x += 1;
                }
            }

            if binds.cw_rot.iter().any(|v| *v == key) {
                for kick in tetrimino.wallkick_table().for_rotation_cw(tetrimino.rotation) {
                    tetrimino.rotate_cw();
                    if check_tetrimino_valid(tetrimino.into(), board, x + kick.0, y - kick.1) {
                        needs_refresh = true;
                        x += kick.0;
                        y -= kick.1;
                        break;
                    } else {
                        tetrimino.rotate_ccw();
                    }
                }
            }

            if binds.ccw_rot.iter().any(|v| *v == key) {
                for kick in tetrimino.wallkick_table().for_rotation_ccw(tetrimino.rotation) {
                    tetrimino.rotate_ccw();
                    if check_tetrimino_valid(tetrimino.into(), board, x + kick.0, y - kick.1) {
                        needs_refresh = true;
                        x += kick.0;
                        y -= kick.1;
                        break;
                    } else {
                        tetrimino.rotate_cw();
                    }
                }
            }

            if binds.soft_drop.iter().any(|v| *v == key) {
                if check_tetrimino_valid(tetrimino.into(), board, x, y + 1) {
                    needs_refresh = true;
                    y += 1;
                    next_fall = Instant::now() + Duration::from_millis(500);
                } else {
                    lock_tetrimino(tetrimino, &mut board, x, y);

                    reset!(next_fall, tetrimino, y, x, hold_used, rando, board);
                }
            }

            if binds.hard_drop.iter().any(|v| *v == key) {
                while check_tetrimino_valid(tetrimino.data(), board, x, y + 1) {
                    y += 1;
                }
                
                lock_tetrimino(tetrimino, &mut board, x, y);

                reset!(next_fall, tetrimino, y, x, hold_used, rando, board);
            }

            if binds.hold.iter().any(|v| *v == key) {
                if !hold_used {
                    tetrimino.rotation = Rotation::Spawn;
                    let held = hold.replace(tetrimino);
                    if let Some(held) = held {
                        next_fall = Instant::now();
                        tetrimino = held;
                        y = 18;
                        x = 3;
                    } else {
                        #[allow(unused_assignments)]
                        {
                            reset!(next_fall, tetrimino, y, x, hold_used, rando, board);
                        }
                    }
                    hold_used = true;
                }
            }
        }

        if Instant::now() > next_fall {
            if check_tetrimino_valid(tetrimino.into(), board, x, y + 1) {
                needs_refresh = true;
                y += 1;
                next_fall += Duration::from_millis(500);
            } else {
                lock_tetrimino(tetrimino, &mut board, x, y);

                reset!(next_fall, tetrimino, y, x, hold_used, rando, board);
            }
        }

        if needs_refresh {
            for y in 0..=39 {
                if board[y].iter().all(Option::is_some) {
                    for y2 in (1..=y).into_iter().rev() {
                        board[y2] = board[y2 - 1];
                    }
                    board[0] = [None; 10];
                }
            }

            let mut ghost_y = y;
            while check_tetrimino_valid(tetrimino.into(), board, x, ghost_y + 1) {
                ghost_y += 1;
            }

            window.clear();
            window.color_set(COLOR_WHITE);
            window.mvprintw(0, 0, include_str!("./outline.txt"));
            print_board(board, &window, 12, -20);
            window.color_set(COLOR_WHITE);
            write_tetrimino_ghost(tetrimino.into(), &window, 2 * x + 12, ghost_y - 20);
            
            write_tetrimino(tetrimino, &window, 2 * x + 12, y - 20);
            if let Some(hold) = hold {
                write_tetrimino_centered(hold, &window, 1, 1);
            }

            rando.reserve(6);
            
            write_tetrimino_centered(rando.remaining()[0], &window, 35, 1);
            write_tetrimino_centered(rando.remaining()[1], &window, 35, 6);
            write_tetrimino_centered(rando.remaining()[2], &window, 35, 9);
            write_tetrimino_centered(rando.remaining()[3], &window, 35, 12);
            write_tetrimino_centered(rando.remaining()[4], &window, 35, 15);
            write_tetrimino_centered(rando.remaining()[5], &window, 35, 18);

            window.refresh();

            window.color_set(COLOR_WHITE);
        }

        thread::sleep(Duration::from_millis(10));
    }

    endwin();
}